#!perl -w

use strict;
use Cwd;
use GD;

my $laz;
my $merged;
my $FORCE = 0;
my $CORES = 1;
my $CLIFF = 45;
while ($_ = shift) {
	if (/^-force$/i) {
		$FORCE = 1;
	}
	elsif (/^-cores$/i) {
		$CORES = shift;
		$CORES = 1 unless ($CORES =~ /^\d+$/ && $CORES > 1);
	}
	elsif (/^-cliff$/i) {
		$CLIFF = shift;
		$CLIFF = 45 unless ($CLIFF =~ /^\d+$/ && $CLIFF > 0);
	}
	elsif (!defined($laz)) {
		$laz = $_;
	}
	elsif (!defined($merged)) {
		$merged = $_;
	}
}

my @files = myglob(split(/,/,$laz));

if (scalar(@files) == 1 && ! ($files[0] =~ /\.la[sz]$/i)) { # single non-LA[SZ] file, use as List of Files:
	open(LOF, $files[0]);
	@files = ();
	while (<LOF>) {
		chomp;
		push(@files, $_);
	}
	close(LOF);
}

die("Usage: gen-slope-asc.pl lazfile|file_with_list_of_laz_files [output_file]\n") unless (scalar(@files));

# Pick up current directory name, use as default output base:
if (! defined($merged)) {
	$merged = getcwd;
	if ($merged =~ /^(.+)[\\\/]([^\\\/]+)$/) {
		$merged = $2;
	}
}

my $LIMIT = 1250000;

my %split = ();

my @slp = ();
my @ele = ();

my $tid = 1;	# This is the main/only thread!
my $CID = "";
my @core_merge = ();

sub generate_files
{
	my ($path) = @_;
	my $perlbase = substr($path,0,length($path)-4);
	
	# Convert to backwards directory slashes for lastools binaries:
	$path =~ s/\//\\/g;
	my $res = `lasinfo -stdout $path`;
	my $base = substr($path,0,length($path)-4);
	my $use_tile_bb = "-use_tile_bb";
	if ($path =~ /tile_(-?\d+)_(-?\d+)_(\d+)\D/) {
		$use_tile_bb = sprintf("-ll %d %d -ncols %s -nrows %s", $1, $2, $3, $3);
	}
	$res =~ s/\n/\t/g;
	my $cmd = sprintf('las2dem -i "%s" -keep_class 2 -extra_pass -slope -set_min_max 0 90 %s -o %s_slope.asc 2>nul', 
		$path, $use_tile_bb, $base); #-invert_ramp 
	$res = `$cmd` unless ($FORCE == 0 && -f $base."_slope.asc");
	push(@slp, $perlbase."_slope.asc");
	
	$cmd = sprintf('las2dem -i "%s" -keep_class 2 -extra_pass -elevation %s -o %s_ele.asc 2>nul', 
		$path, $use_tile_bb, $base);
	$res = `$cmd` unless ($FORCE == 0 && -f $base."_ele.asc");
	push(@ele, $perlbase."_ele.asc");
}	

my ($XMIN, $XMAX, $YMIN, $YMAX, $STEP) = find_min_max(@files);

my @tids = ();

if ($CORES > 1) { # Start sub-processes!
	for (my $core = 0; $core < $CORES; $core++) {
		my @sub_tiles = ();
		for (my $f = $core; $f < scalar(@files); $f += $CORES) {
			push(@sub_tiles, $files[$f]);
			my $perlbase = substr($files[$f],0,length($files[$f])-4);
			push(@slp, $perlbase."_slope.asc");
			push(@ele, $perlbase."_ele.asc");
		}
		$tid = fork;
		if ($tid == 0) { # Sub-thread!
			@tids = ();
			@files = @sub_tiles;
			$CID = "$core: ";
			last;
		}
		printf(STDERR "PID $tid forked\n");
		push(@tids, $tid);
	}
}

if ($tid == 0 || $CORES <= 1) { # Process the list of files!
	my $cnt = 0;
	foreach (@files) {
		#next if (/^\.+$/);
		next unless (/tile_([\-]?\d+)_([\-]?\d+)\D/);
		
		my ($x, $y) = ($1, $2);
		my $path = $_;
		printf(STDERR "$CID%d/%d: $path\n", ++$cnt, scalar(@files));
		generate_files($path);
	}
	if ($tid == 0) { # Is this a sub-thread?
		die("$CID Done\n");
	}
}
else {
	my $kid;
	foreach (@tids) {
		$kid = waitpid($_, 0);
		#sleep(1);
	}
}

@tids = ();
my $core = 3;
if ($CORES > 1) {
	$core = 1;
	$tid = fork;
	if ($tid == 0) { # Sub-thread!
		@tids = ();
		$CID = "0: ";
	}
	else {
		printf(STDERR "PID $tid forked\n");
		push(@tids, $tid);
		$core = 2;
		$tid = fork;
		if ($tid == 0) { # Sub-thread!
			@tids = ();
			$CID = "1: ";
		}
		else {
			printf(STDERR "PID $tid forked\n");
			push(@tids, $tid);
			$core = 0;
		}
	}
}

my $outfile;

if ($core & 1) {
	$outfile = $merged."_slope.asc";

	printf(STDERR "\n%sMerging slope files\n", $CID);
	my ($xsize, $ysize, $x0, $y0, $cell, $NODATA, @img) = merge_asc($STEP,@slp);
	printf(STDERR "%sWriting merged slope file %s\n", $CID, $outfile);
	write_asc($outfile, $xsize, $ysize, $x0, $y0, $cell, $NODATA, @img);

	my ($xmin, $ymin, $xmax, $ymax) = ($x0, $y0, $x0+$xsize, $y0+$ysize);

	printf(STDERR "%sCreating slope and cliff files.\nSize: (%d,%d)-(%d-%d)\n", $CID,$xmin,$ymin,$xmax,$ymax);

	my $img = GD::Image->new($xsize, $ysize);
	my @hist = (0) x 91;
	for (my $y = 0; $y < $ysize; $y++) {
		$_ = $img[$y];
		$_ = '' unless defined($_);
		my @val = split;
		for (my $x = 0; $x < $xsize; $x++) {
			my $a = $val[$x];
			$a = 0 unless (defined($a) && $a >= 0);
			$img->setPixel($x, $y, $a);
			$hist[$a]++;
		}
		printf(STDERR "%d%%\r", $y*100/$ysize) if ($core == 3);
	}

	printf(STDERR "%sWriting %s_slope.gif\n", $CID,$merged);
	write_slope_gif($merged."_slope.gif", $img, 45, 1, $xmin, $ymax);

	printf(STDERR "%sWriting %s_cliff.gif\n", $CID,$merged);
	write_cliff_gif($merged."_cliff.gif", $img, $CLIFF, 90, 1, $xmin, $ymax);
	#printf(STDERR "%sWriting %s_cliff1.gif\n", $CID,$merged);
	#write_cliff_gif($merged."_cliff1.gif", $img, 45, 63, 1, $xmin, $ymax);
	#printf(STDERR "%sWriting %s_cliff2.gif\n", $CID,$merged);
	#write_cliff_gif($merged."_cliff2.gif", $img, 64, 71, 1, $xmin, $ymax);
	#printf(STDERR "%sWriting %s_cliff3.gif\n", $CID,$merged);
	#write_cliff_gif($merged."_cliff3.gif", $img, 72, 90, 1, $xmin, $ymax);

	die("Slope/Cliff processing done!\n") if ($core == 1);
}
if ($core & 2) {

	printf(STDERR "%sMerging elevation files\n", $CID);
	$outfile = $merged."_elevation.asc";

	write_asc($outfile, merge_asc($STEP, @ele));

	die("Elevation/DEM processing done!\n") if ($core == 2);
}
if ($core == 0) {
	foreach (@tids) {
		my $kid = waitpid($_, 0);
	}
}

printf(STDERR "\nAll done!");

# The power value (0.7) exaggerates low slope angles

sub power
{
	my ($x, $p) = @_;
	my $res = 0;
	if ($x > 0) {
		$res = exp(log($x)*$p);
	}
	elsif ($x < 0) {
		$res = -exp(log(-$x)*$p);
	}
	return $res;
}

sub write_slope_gif
{
	my ($outfile, $img, $full_black, $CELL, $xmin, $ymax) = @_;
	
	for (my $c = $img->colorsTotal; $c > 0; $c--) {
		$img->colorDeallocate($c-1);
	}
	for (my $a = 0; $a <= 90; $a++) {
		my $grey = 0;
		if ($a < $full_black) {
			$grey = int(255*(1-power($a/$full_black,0.7))+0.5);
		}
		$img->colorAllocate($grey, $grey, $grey);
	}

	open(D, ">$outfile");
	binmode(D);
	print D $img->gif;
	close D;

	my $world = substr($outfile,0,length($outfile)-3).'gfw';
	open(W, ">$world") or die "$!";
	foreach ($CELL, 0, 0, -$CELL, 
		 $xmin + 0.5,
		 $ymax - 0.5
		 ) {
		printf(W "%1.4f\n", $_);
	}
	close(W);
}

sub write_cliff_gif
{
	my ($outfile, $slope, $min_black, $max_black, $CELL, $xmin, $ymax) = @_;
	
	my ($sizex, $sizey) = $slope->getBounds;
	my $cliff = GD::Image->new($sizex, $sizey);
	
	my $white = $cliff->colorAllocate(255,255,255);
	my $black = $cliff->colorAllocate(0,0,0);
	my @tab = ();
	for (my $a = 0; $a <= 255; $a++) {
		my $color = $white;	
		if ($a >= $min_black && $a <= $max_black) {
			$color = $black;
		}
		$tab[$a] = $color;
	}
	
	$cliff->fill(0,0,$white);
	
	for (my $y = 0; $y < $sizey; $y++) {
		for (my $x = 0; $x < $sizex; $x++) {
			my $a = $slope->getPixel($x, $y);
			$a = $tab[$a];
			if ($a == $black) {
				$cliff->arc($x,$y,2,2,0,360, $black);
			}
			#$cliff->setPixel($x, $y, $a);
		}
	}

	open(D, ">$outfile");
	binmode(D);
	print D $cliff->gif;
	close D;

	my $world = substr($outfile,0,length($outfile)-3).'gfw';
	open(W, ">$world") or die "$!";
	foreach ($CELL, 0, 0, -$CELL, 
		 $xmin + 0.5,
		 $ymax - 0.5
		 ) {
		printf(W "%1.4f\n", $_);
	}
	close(W);
}

sub merge_asc
{
	my ($STEP, @asc) = @_;
	
	my ($xmin, $ymin, $xmax, $ymax) = (1e38,1e38,-1e38,-1e38);

	my @sorted_asc = ();
	foreach (@asc) {
		my ($x0, $y0, $x1, $y1, $nr, $nc);
		if (/tile_([\-]?\d+)_([\-]?\d+)_(\d+)\D/) {
			($x0, $y0, $STEP) = ($1,$2,$3);
			$nr = $nc = $STEP;
		}
#		elsif (/split.*_(\d{6})_(\d{7})/) {
#			$nr = $nc = $STEP*0.5;
#			($x0, $y0) = ($1,$2);
#		}
		else {
			die("Unknown file! $_");
		}
		($x1, $y1) = ($x0+$nc,$y0+$nr);
		$xmin = $x0 if ($x0 < $xmin);
		$ymin = $y0 if ($y0 < $ymin);
		$xmax = $x1 if ($x1 > $xmax);
		$ymax = $y1 if ($y1 > $ymax);
		push(@sorted_asc, sprintf("%7d,%6d,%4d,%4d,%s", 9999999-$y0, $x0, $nc, $nr, $_));
	}
	printf(STDERR "Size: (%d,%d)-(%d-%d)\n", $xmin,$ymin,$xmax,$ymax);
	
	my @img;
	my ($nrows, $ncols, $xll, $yll, $cell, $NODATA);
	my $blank;
	my $cnt = 0;
	foreach (sort @sorted_asc) {
		#if (/442250_6619750/) {
		#	printf(STDERR "Should be ok...\n");
		#}
		my ($y0, $x0, $nc, $nr, $fname) = split(/,/);
		printf(STDERR "%d/%d %s\n", ++$cnt, scalar(@sorted_asc), $fname);
		$y0 = 9999999 - $y0;
		
		my %tags = ();

		open(S, $fname) || next; # Skip any missing files!
		my $hdr = 0; 
		while (<S>) {
			chomp; 
			last if (/^\s*$/);		# Blank line?
			my ($tag, $val) = split;
			$tags{$tag} = $val;
			last if (++$hdr >= 6);
		}
		next unless ($hdr); # Empty file, just skip it without any error messages!

		$ncols = $tags{ncols};
		$nrows = $tags{nrows};
		$xll = $tags{xllcorner};
		$yll = $tags{yllcorner};
		$cell = $tags{cellsize};
		$NODATA = $tags{NODATA_value};
		#die("Coordinate mismatch in $fname: $x0 - $xll, $y0 - $yll\n") 
		unless (defined($xll) && defined($yll) && defined($xll) && defined($yll) && defined($cell)) {
			printf(STDERR "Missing tags in $fname\r");
			next;
		}
		unless (($x0 <= $xll) && ($x0+256 >= $xll+$ncols) && 
			    ($y0 <= $yll) && ($y0+256 >= $yll+$nrows)) {
			warn("Header error in $fname\r");
			next;
		}
		my ($xoff, $yoff) = ($xll-$xmin, $ymax-($yll+$nrows));
		$blank = join(' ', ($NODATA) x ($xmax-$xmin)) unless defined($blank);
		
		while (<S>) {
			chomp;
			#s/$NODATA/0/g;
			my @val = split;
			die("Wrong number of columns: $ncols\n") unless ($ncols == scalar(@val));
			
			$_ = $img[$yoff];
			if (!defined($_)) {
				$_ = $blank;
			}
			my @line = split;
			splice(@line, $xoff, $ncols, @val);
			$img[$yoff] = join(' ', @line);
			$yoff++;
		}
		close(S);
	}
	#($ncols, $nrows, $xll, $yll, $cell, 0, @img)
	return ($xmax-$xmin, $ymax-$ymin, $xmin, $ymin, $cell, $NODATA, @img);
}

sub write_asc
{
	my ($fname, $ncols, $nrows, $xll, $yll, $cell, $NODATA, @img) = @_;
	
	open(my $fh, '>', $fname) || die("Could not write to $fname\n");
	printf($fh "ncols $ncols\nnrows $nrows\nxllcorner $xll\nyllcorner $yll\ncellsize $cell\nNODATA_value $NODATA\n");
	foreach (@img) {
		printf($fh "%s\n", $_);
	}
	close($fh);
}

sub myglob
{
	my @match = ();
	foreach (@_) {
		my $dir = '.';
		if (/^(.+)[\\\/]([^\\\/]+)$/) { # Starts with a directory!
			$dir = $1;
			$_ = $2;
		}
		s/\./\t/g;		# Save away any explicit '.' symbols
		s/\*/\.*/g;		# Convert * to .*
		s/\?/\./g;		# Convert ? to .
		s/\t/\\\./g;	# Restore explicit '.' as escaped '\.'
		
		my $pattern = $_;
		
		opendir(D, $dir);
		my @files = readdir(D);
		closedir(D);
		
		$dir = ($dir eq '.') ? '' : $dir .'/';
		
		foreach (@files) {
			next unless (-f $dir.$_);
			
			next unless (/^$pattern$/);
			push(@match, $dir . $_);
		}
	}
	return @match;
}

sub deltas
{
	my @steps = sort {$a <=> $b} @_;
	return (undef, undef) if (scalar(@steps) <= 1);

	my ($min, $max) = (1e38, 0);
	for (my $i = 1; $i < scalar(@steps); $i++) {
		my $step = $steps[$i]-$steps[$i-1];
		$min = $step if ($step < $min);
		$max = $step if ($step > $max);
	}
	return ($min, $max);
}

sub find_min_max
{
	my @files = @_;
	
	my ($xmin, $xmax, $ymin, $ymax, $adaptive) = (1e38,-1e38,1e38,-1e38, 0);
	if (scalar(@files) > 1) {
		# New & even faster method: tile_EEEEE_NNNNN_SSS.laz naming for adaptive tile sizes:
		
		my ($smin, $smax) = (1e38, 0);
		foreach (@files) {
			goto skip unless (/tile_([\-]?\d+)_([\-]?\d+)_(\d+)\D/);

			my ($e, $n, $s) = ($1, $2, $3);
			$xmin = $e if ($e < $xmin);
			$xmax = $e if ($e > $xmax);
			$ymin = $n if ($n < $ymin);
			$ymax = $n if ($n > $ymax);
		}
		$adaptive = -1;
		return ($xmin, $xmax, $ymin, $ymax, $adaptive);
		
skip:
		# Try to use fast method: Look for tile_EEEEEE_NNNNNNN.laz names!
		my %ehist = ();
		my %nhist = ();
		my $fast = 1;
		foreach (@files) {
			unless (/tile_([\-]?\d+)_([\-]?\d+)\D/) {
				$fast = 0;
				last;
			}
			$ehist{$1}++;
			$nhist{$2}++;
		}
		if ($fast) {
			my @ehist = sort {$a <=> $b} keys %ehist;
			my @nhist = sort {$a <=> $b} keys %nhist;
			my ($min, $max) = deltas(@ehist);
			my $step = $min;
			($min, $max) = deltas(@nhist);
			if ($min == $step) {
				$step = 256 if ($step < 256);
				
				$xmin = $ehist[0];
				$xmax = $ehist[-1] + $step;
				$ymin = $nhist[0];
				$ymax = $nhist[-1] + $step;
					
				return ($xmin, $xmax, $ymin, $ymax, $step);
			}
		}
	}
	my $ground_points = 0;
	foreach (@files) {
		chomp;
		my $path = $_;
		$path =~ s/\//\\/g;
		my $res = `lasinfo -stdout $path`;
		if (0) { #($res =~ /LAStiling .+ bbox (\S+)\s(\S+)\s(\S+)\s+(\S+), buffer/) {
			$xmin = $1 if ($1 < $xmin);
			$ymin = $2 if ($2 < $ymin);
			$xmax = $3 if ($3 > $xmax);
			$ymax = $4 if ($4 > $ymax);
		}
		else {
			die("Not a LAS/LAZ file? $_\n") unless ($res =~ /min x y z:\s+(\S+)\s+(\S+)\s+(\S+)/);
			$xmin = $1 if ($1 < $xmin);
			$ymin = $2 if ($2 < $ymin);
			die("Not a LAS/LAZ file? $_\n") unless ($res =~ /max x y z:\s+(\S+)\s+(\S+)\s+(\S+)/);
			$xmax = $1 if ($1 > $xmax);
			$ymax = $2 if ($2 > $ymax);
		}
		my $g = 0;
		if ($res =~ /\s(\d+)\s+Ground (2)/) {
			$g = $1;
		}
		$ground_points += $g;
	}
	#printf(STDERR "Raw X: %1.2f - %1.2f\nY: %1.2f - %1.2f\n", $xmin, $xmax, $ymin, $ymax);
	
	$xmin = adjust($xmin, 50);
	$ymin = adjust($ymin, 50);
	$xmax = adjust($xmax, -50);
	$ymax = adjust($ymax, -50);
	
	return ($xmin, $xmax, $ymin, $ymax);
}

sub adjust
{
	my ($x, $adj) = @_;
	$x = int($x+0.5); 		# Round to nearest integer!
	my $ax = $x+$adj; 		# Round to nearest integer!
	$x = $ax if (($ax % 100) == 0);
	return $x;
}

