#! perl -w

# Convert AR5 data to DXF!
# Updated 2018-12-11, including Sverre Foyen's GDAL compatibility update for areas.
# Update 2020-01-02: Hack to convert Oslo 'Mast' symbols with zero-length lines into POINTs

use strict;
use Cwd;
use Encode qw( encode decode );
use IO::Uncompress::Unzip;
use Geo::Coordinates::UTM;
use File::stat;

my $DIST_SCALE = 0.5;

my $TELLEKURVEEKVIDISTANSE = 25;
my $EKVIDISTANSE = 5;

my %ignore_sosi = ('ArealressursGrense' => 1,'flate_ARF_Skog_areal'=> 1, 
		'flate_ARF_Vann'=> 1, 'ArealressursGrenseFiktiv'=> 1, 'flate_ArealressursFlate'=>1,
		'flate_ARF_Hav' => 1, 'flate_ARF_Samferdsel' => 1, 'flate_ARF_Ferskvann' => 1,
		'Monelinje' => 1, 'Taksprang' => 1, 'Bygningslinje' => 1, 'TakSprangBunn' => 1, 'Bygningsdelelinje' => 1);

my $sos;
my $dxf;
my $VERBOSE = 0;
my $DATE = "1970-01-01";
my $CORES = 0;
my $TARGETUTM = "";
my $KOMMUNENR = "";
my $SOURCEUTM = "";
my $CRT = 1;
my $AUTODIR = 0; # Generate directory based on kommune nr/navn
my $UPDATE = 0; # Set to 1 to use MAKE logic, i.e. check timestamps
my $ODIR = '';

# Which rectangle should we include? Default to everything!
my ($XMIN, $YMIN, $XMAX, $YMAX) = (-1e38,-1e38,1e38,1e38);

while ($_ = shift) {
	if (/^-verbose$/i) {
		$VERBOSE = 1;
	}
	elsif (/^-autodir$/i) {
		$AUTODIR = 1;
	}
	elsif (/^-odir$/i) {
		$ODIR = shift;
	}
	elsif (/^-update$/i) {
		$UPDATE = 1;
	}	
	elsif (/^-date$/i) {
		$_ = shift;
		$DATE = $_ if ($_ =~ /^\d+-\d\d-\d\d$/);
	}
	elsif (/^-cores$/i) {
		$_ = shift;
		$CORES = $_ if ($_ =~ /^\d+$/);
	}
	elsif (/^-crt$/i) {
		$CRT = shift;
	}
	elsif (/^-target_utm$/i) {
		$_ = shift;
		$TARGETUTM = uc($_) if ($_ =~ /^\d+.?$/);
		printf(STDERR "Target UTM = %s\n", $TARGETUTM);
	}
	elsif (/^-kommune_nr$/i) {
		$_ = shift;
		$KOMMUNENR = $_ if ($_ =~ /^\d+$/);
		printf(STDERR "Kommune nr = %s\n", $KOMMUNENR);
	}
	elsif (/^-ekvidistanse$/i) {
		$_ = shift;
		if (/^(\d+)[-:,](\d+)$/) {
			($EKVIDISTANSE, $TELLEKURVEEKVIDISTANSE) = ($1, $2);
		}
		elsif (/^\d+$/) {
			($EKVIDISTANSE, $TELLEKURVEEKVIDISTANSE) = ($_, $_*5);
		}
	}
	elsif (/^-keep_xy$/i) {
		$XMIN = shift; $YMIN = shift;
		$XMAX = shift; $YMAX = shift;
	}
	elsif (!defined($sos)) {
		$sos = $_;
	}
#	elsif (!defined($dxf)) {
#		$dxf = $_;
#	}
}

$CORES = 1 if ($CORES < 1);

if ($KOMMUNENR) {}

my @files;
if (defined($sos)) {
	@files = myglob(split(/,/,$sos));
}
elsif ($KOMMUNENR) {
	opendir(D, ".");
	my @dirs = readdir(D);
	closedir(D);
	foreach (@dirs) {
		if (/^(FKB|Matrikkelen)\s?-\s?\S.+\s+SOSI-format$/ ) {
			my $dir = $_;
			opendir(D, $dir);
			my @f = readdir(D);
			closedir(D);
			my @match = ();
			my %found = ();
			foreach (@f) {
				if (/^Basisdata_(\d\d\d\d)_(\S.+)_(\d+)_(.+)_SOSI.zip/i) {
					my ($knr, $knavn, $proj, $stype) = ($1, $2, $3, $4, $5);
					if ($knr eq $KOMMUNENR) {
						push(@match, $dir.'/'.$_);
						$found{$proj}++;
					}
				}
			}
			if (scalar(@match) > 1 && scalar(keys %found) > 1 && $TARGETUTM) {
				my $m = "258".substr($TARGETUTM,0,2);
				if (!defined($found{$m})) {
					$m = 5940 + substr($TARGETUTM,0,2);
				}
				foreach (@match) {
					if (!defined($found{$m}) || (index($_, $m) >= 0)) {
						push(@files, $_);
					}
				}
			}
			else {
				push(@files, @match);
			}
		}
	}
}

if (scalar(@files) == 1 && ! ($files[0] =~ /\.sos[i]?$/i)
	&& ! ($files[0] =~ /\.zip$/i)) { # single non-SOS(i)/non-ZIP file, use as List of Files:
	open(LOF, $files[0]);
	@files = ();
	while (<LOF>) {
		chomp;
		push(@files, $_);
	}
	close(LOF);
}

die("Usage: sosi_2_dxf.pl [options] sosifile(s)|file_with_list_of_sosi_files [output file]\n\
Options:\n\t-cores N\n\t-verbose\n\t-date yyyy-mm-dd\n\t-keep_xy xmin ymin xmax ymax\n
\t-crt 0|1\n\t-target_utm xxN\n\t-autodir\n\t-kommune_nr nnnn\n") 
unless (scalar(@files));

printf(STDERR "Processing %d files:\n%s\n", scalar(@files), join("\n", @files));

my %crt = ('form_line' => 103, 'contour' => 101, 'index_contour' => 102,
'AnnenBygning' => 526, 'AnnetGjerde' => 522, 'AnnetVegarealAvgrensning' => 529.001, 'Arealbrukgrense' => 414,
'BeskrivendeHjelpelinjeAnlegg' => 529.002, 'BronnGrense' => 301.001, 'Bru' => 0, 'Bruavgrensning' => 529.003,
'Bygning' => 526, 'Bygningslinje' => 853, 'Dam' => 526, 'Damkant' => 309.001, 'flate_ElvBekk' => 305,
'ElvBekkKant' => 301.001, 'ElveElvSperre' => 301.002, 'ElvelinjeFiktiv' => 0, 'Flaggstang' => 539,
'FrittstaendeTrappKant' => 861.005, 'GangSykkelveg' => 505, 'GangSykkelvegkant' => 529.004,
'Grunnmur' => 861.006, 'Grustak' => 211, 'Hekk' => 0, 'Idrettsanlegg' => 401, 'flate_Innsjo' => 301, 'InnsjoElvSperre' => 301.002,
'Innsjokant' => 301.001, 'InnsjokantRegulert' => 301.001, 'InnsjolinjeFiktiv' => 301.002, 'KaiBrygge' => 0,
'KaiBryggeKant' => 301.001, 'KanalGroft' => 306,'flate_KanalGroft' => 302, 'KantUtsnitt' => 0, 'Lavebru' => 0, 'Monelinje' => 0, 'MurLoddrett' => 861.006,
'ParkeringsomradeAvgrensning' => 529.005, 'Stein' => 206, 'SteinOmriss' => 203.001, 'Steingjerde' => 519, 'Sti' => 506,
'Stikkrenne' => 306, 'Takkant' => 853, 'Taksprang' => 0, 'Tank' => 0, 'Tankkant' => 853, 'Traktorveg' => 505,
'Traktorvegkant' => 529.006, 'TrappBygg' => 0, 'Veg' => 504, 'Vegdekkekant' => 529.001, 'VegkantAnnetVegareal' => 529.001,
'VegkantAvkjorsel' => 529.001, 'VegkantFiktiv' => 529.001, 'VegoppmerkingLangsgaende' => 0, 'Vegrekkverk' => 522,
'Vegsperring' => 0, 'Veranda' => 0, 'Alpinbakke' => 0, 'Anleggsomrade' => 0, 'ArealbrukgrenseFiktiv' => 0,
'BygningBru' => 0, 'Bygningsdelelinje' => 861.005, 'Campingplass' => 0, 'FeristAvgrensning' => 0, 'FiktivAvgrensningForAnlegg' => 0,
'FiktivBygningsavgrensning' => 0, 'Flytebrygge' => 0, 'Fortauskant' => 0, 'FortauskantYtre' => 0, 'FrittstaendeTrapp' => 0,
'Fundament' => 0, 'FundamentKant' => 0, 'GangfeltAvgrensning' => 0, 'Gravplass' => 401, 'Hjelpelinje3D' => 0, 'KanalGroftKant' => 0,
'Kjorebanekant' => 529.001, 'Lekeplass' => 0, 'Lysloype' => 505, 'MurFrittstaende' => 0, 'Parkeringsomrade' => 529.001, 'Skitrekk' => 0,
'Skjerm' => 0, 'SkraForstotningsmur' => 861.006, 'SkraForstotningsmurAvgrensning' => 861.006, 'SportIdrettPlass' => 0, 'Stolheis' => 516,
'Svommebasseng' => 302, 'SvommebassengKant' => 309.001, 'Takoverbygg' => 0, 'TakoverbyggKant' => 0, 'Takplata' => 0, 'TaksprangBunn' => 0,
'Trafikkoy' => 0, 'Trafikkoykant' => 529.001, 'VeggFrittstaende' => 0, 'VeggroftApen' => 0, 'Vegskulderkant' => 0, 'Voll' => 0,
'Bronn' => 312,
'Brudetalj' => 309.001,
'FiktivDelelinje' => 0,
'FlytebryggeKant' => 309.001,
'Fordelingsskap' => 0,
'Gangvegkant' => 0,
'Golfbane' => 415,
'HavElvSperre' => 301.002,
'flate_Havflate' => 301,
'Hjelpepunkt3D' => 0,
'IndustriOmrade' => 0,
'InnmaltTre' => 418,
'Kystkontur' => 309.001,
'KystkonturTekniskeAnlegg' => 309.001,
'LuftledningHSP' => 516,
'MastElGr' => 516.001,
'MastTele' => 516.001,
'MastTeleGr' => 516.001,
'MastVeilys' => 516.001,
'Molo' => 0,
'MoloKant' => 309.001,
'NettstasjonGr' => 0,
'NettstasjonMast' => 516.001,
'PblTiltak' => 0,
'PblTiltakGrense' => 0,
'RbFormalOmrade' => 0,
'RpArealformalOmrade' => 0,
'RpFormalGrense' => 0,
'SOSI_Objekt' => 0,
'Silo' => 0,
'Silokant' => 0,
'Skjar' => 0,
'Slipp' => 540,
'Steinbrudd' => 0,
'Steintipp' => 0,
'Sti' => 507,
'StolpeEnkel' => 516.001,
'StolpeStor' => 516.001,
'Tarn' => 536,
'Tarnkant' => 0,
'Tribune' => 0,
'Tribunekant' => 0,
'Tunnelportal' => 518,
'Forsenkningskurve' => 101.999,
'Forsenkningspunkt' => 115,
'Hjelpekurve' => 103,
'Tellekurve' => 102,
'Hoydekurve' => 101,
'Terrengpunkt' => 112,
'Toppunkt' => 112,
'ArealressursFlate' => 0,
'ArealressursGrense' => 0,
'ArealressursGrenseFiktiv' => 0,
'BautaStatue' => 0,
'BautaStatueGrense' => 0,
'Bensinpumpe' => 0,
'Dataavgrensning' => 0,
'Elveterskel' => 0,
'ElveterskelKant' => 0,
'FlomlopKant' => 0,
'Flytebryggekant' => 0,
'Kulvert' => 0,
'Kum' => 0,
'MastElFagverk' => 0,
'Rasteplass' => 0,
'Rorgate' => 0,
'Sluk' => 0,
'Tommervelte' => 0,
'Elveforbygningskant' => 0,
'Fasadeliv' => 0,
'Flotningsrenne' => 0,
'Hoppbakke' => 0,
'Hydrant' => 0,
'NettstasjonKiosk' => 0,
'OppdrettsmerderKant' => 0,
'Portstolpe' => 0,
'Referansestolpe' => 0,
'Skiltportal' => 0,
'SnoIsbre' => 0,
'SnoIsbreKant' => 0,
'Myr' => 310,
'Skog' => 406,
'ApentOmrade' => 403,
'DyrketMark' => 415,
'FerskvannTorrfall' => 211,
'FerskvannTorrfallkant' => 0,
'Industriomrade' => 529,
'InnsjoInnsjoSperre' => 0,
'TettBebyggelse' => 527,
'Tregruppe' => 419,
'Alpinbakke' => 401,
'Lufthavn' => 529,
'Rullebane' => 529,
'Helikopterplass' => 529,
'Steinbrudd' => 212,
'ARF_Apen_fastmark' => 402,
'ARF_Bebygd_areal' => 527,
'ARF_Dyrket_mark' => 415,
'ARF_Eng' => 401,
'ARF_Jorde' => 403,
'ARF_Myr' => 310,
'ARF_Samferdsel' => 529,
'ARF_Skog_areal' => 0,
'ARF_Vann' => 301,
'flate_GangSykkelveg' => 529,
'flate_Parkeringsomrade' => 529,
'flate_Trafikkoy' => 0,
'flate_Traktorveg' => 0,
'flate_Veg' => 529,
'flate_Vann' => 301);

my %ArType = (10 => 'Bebygd_og_samferdsel', 11 => 'Bebygd_areal', 12 => 'Samferdsel', 
	20 => 'Dyrket_mark', 21 => 'Dyrket_mark', 22 => 'Jorde', 23 => 'Eng', 30 => 'Skog_areal', 50 => 'Apen_fastmark',
	60 => 'Myr', 70 => 'Snoisbre', 80 => 'Vann', 81 => 'Ferskvann', 82 => 'Hav');


my ($xmin, $xmax, $ymin, $ymax) = (1e38, -1e38, 1e38, -1e38);
my ($typ, $nr, $objtype, $objdate, $hoyde, $artype, $arealtype, $flatenr, $ref, 
    $datafangstdato, $verifiseringsdato, $kopidato, @coords) = ();

my %objects = ();
my %seen_crt = (); # All seen object types will end up here for the CRT file!

my ($pe, $pn) = ();
my $objnr;
my %objlist = ();
my ($yy, $mm, $dd) = ();
my $ENHET = 0.01;
my @ORIGO_NE = (0,0);
my $line = 0;
my $nrOfObjects = 0;
my $UTF8_TO_ANSI = 0;


sub flushprev
{
	return unless (defined($objtype) && scalar(@coords));
    if (defined($kopidato)) {$objdate = $kopidato;}
    elsif (defined($verifiseringsdato)) {$objdate = $verifiseringsdato;}
    elsif (defined($datafangstdato)) {$objdate = $datafangstdato;}
    
    if (defined($objdate)) {
        if ($objdate lt $DATE) {
            printf(STDERR "Skipping object: date = $objdate\n") if ($VERBOSE);
            return;
        }
        printf(STDERR "Keeping object with date = $objdate\n") if ($VERBOSE);
    }
	# Sjekk for sty i enkelt-koordinater:
	my @c = ($coords[0]);
	for (my $c = 2; $c < scalar(@coords); $c++) {
		my $sqstep = square_dist2d($coords[$c-2], $coords[$c-1]);
		if ($sqstep > 10000 && $sqstep > 100*square_dist2d($coords[$c-2], $coords[$c])) {
			printf(STDERR "Koordinat med sty, hopper over: %f %f\n", $sqstep, dist2d($coords[$c-2], $coords[$c]));
		}
		else {
			push(@c, $coords[$c-1]);
		}
	}
	push(@c, $coords[-1]);
	@coords = @c;
	
	@{$objlist{$objnr}} = @coords if (defined($objnr));
	if (defined($flatenr)) {
		#@coords = () if scalar()
		@coords = ($objtype, $ref, @coords);
		#@coords = ($objtype, $ref);
		#printf(STDERR "Flate %d type: %s\n", $flatenr, $objtype);
		$objtype = 'flate';
	}
	push(@{$objects{$objtype}}, join(';',@coords));
	($typ,$nr, $flatenr, $ref, $yy, $mm, $dd, $objdate, $datafangstdato, $verifiseringsdato, $kopidato, @coords) = ();
}

sub conv
{
	my ($e, $n, $h) = @_;
	($e, $n) = ($e*$ENHET+$ORIGO_NE[1],$n*$ENHET+$ORIGO_NE[0]);
	if (defined($h)) {
		if ($h ne '') {
			$h *= $ENHET;
		}
		return ($e, $n, $h);
	}
	return ($e, $n);
}

my ($tid, @tids) = (0);

my $CID = '';
my %fsizecache = ();

sub fsize
{
	my ($fname) = @_;
	return $fsizecache{$fname} if (defined($fsizecache{$fname}));
	my $fs = -s $fname;
	#(stat($fname))[5];
	$fsizecache{$fname} = $fs;
	return $fs;
}

if ($CORES > 1) {
	my @allfiles = reverse sort {fsize($a) <=> fsize($b)} @files;
	for (my $core = 0; $core < $CORES; $core++) {
		@files = ();
		$CID = sprintf("%d: ",$core);
		for (my $f = $core; $f < scalar(@allfiles); $f += $CORES) {
			push(@files, $allfiles[$f]);
		}
		$tid = fork;
		last if ($tid == 0); # Sub-thread
		push(@tids, $tid);
		printf(STDERR "%s started\n", $core);
	}
}

if ($tid) {
	foreach (@tids) {
		my $kid = waitpid($_, 0);
		#sleep(1);
	}
	printf(STDERR "All threads finished!\n");
	die("All done!");
}

my $i = 0;
foreach (@files) {
	$i++;
    my $sos = $_;
    my $dxf = substr($_, 0, length($_)-4);
    $dxf = $dxf.'.' unless(substr($dxf,-1) eq '.');
    $dxf = $dxf.'dxf';
	if ($AUTODIR && $dxf =~ /Basisdata_(\d+_\S+)_(\d+)_FKB-(\S+)_SOSI\.dxf$/i) {
		my ($knr_navn, $proj, $typ) = ($1, $2, $3);
		$knr_navn = $TARGETUTM.$knr_navn;
		mkdir($knr_navn);
		$dxf = $knr_navn."/".$knr_navn.'_FKB-'.$typ.'_SOSI.dxf';
	}
	elsif ($AUTODIR && $dxf =~ /Basisdata_(\d+_\S+)_(\d+)_Matrikkelen(\S+)_SOSI\.dxf$/i) {
		my ($knr_navn, $proj, $typ) = ($1, $2, $3);
		$knr_navn = $TARGETUTM.$knr_navn;
		mkdir($knr_navn);
		$dxf = $knr_navn."/".$knr_navn.'_Matrikkelen-'.$typ.'_SOSI.dxf';
	}

	if ($ODIR && $dxf =~ /Basisdata_(\d+_\S+)_(\d+)_FKB-(\S+)_SOSI\.dxf$/i) {
		my ($knr_navn, $proj, $typ) = ($1, $2, $3);
		$knr_navn = $TARGETUTM.$knr_navn;
		$dxf = $ODIR."/".$knr_navn.'_FKB-'.$typ.'_SOSI.dxf';
	}
	elsif ($ODIR && $dxf =~ /Basisdata_(\d+_\S+)_(\d+)_Matrikkelen(\S+)_SOSI\.dxf$/i) {
		my ($knr_navn, $proj, $typ) = ($1, $2, $3);
		$knr_navn = $TARGETUTM.$knr_navn;
		$dxf = $ODIR."/".$knr_navn.'_FKB-'.$typ.'_SOSI.dxf';
	}

	if ($UPDATE) {
		# Check to see if the output file already exists, is non-empty and is newer than the source:
		if (-e $dxf) {
			my $src = stat($sos);
			my $dst = stat($dxf);
			if (defined($src) && defined($dst) && $dst->size && $dst->mtime >= $src->mtime) {
				printf(STDERR "$CID Skipping %s\n", $dxf);
				next; # Skipping, no update needed!
			}
		}
	}
	printf(STDERR "$CID %d/%d %s->%s\n", $i, scalar(@files), $sos, $dxf);
	my $sf;
	if ($sos =~ /\.zip$/i) {
		$sf = new IO::Uncompress::Unzip($sos) || die("Could not unpack $sos");
	}
	else {
		open($sf, '<', $sos) || die("Could not read from $sos\n");
	}
    open(DXF, ">$dxf") || die("Could not write to $dxf\n");

    ($xmin, $xmax, $ymin, $ymax) = (1e38, -1e38, 1e38, -1e38);
    ($typ, $nr, $objtype, $objdate, $hoyde, $artype, $arealtype, $flatenr, $ref, 
        $datafangstdato, $verifiseringsdato, $kopidato, @coords) = ();

    %objects = ();

    ($pe, $pn) = ();
    $objnr = undef;
    %objlist = ();
    ($yy, $mm, $dd) = ();
    $ENHET = 0.01;
    @ORIGO_NE = (0,0);
    $line = 0;
    $nrOfObjects = 0;
    $UTF8_TO_ANSI = 0;

    while (<$sf>) {
        s/[\r\n]*$//;
		#chomp;
        $line++;
        if (/^\./) {
            if ($UTF8_TO_ANSI) {	# Convert from utf8 to ansi if needed
                my $code_points = decode('UTF-8', $_);
                $_ = encode('cp1252', $code_points);
            }
            
            next if (/^\.\.N/);	# Header line before (n,e) coordinates

            flushprev;
            $nrOfObjects++;
            #last if (defined($MaxObjects) && $nrOfObjects >= $MaxObjects);
            
            if (/\.(PUNKT)\s+(\d+):\s*$/  || /\.(KURVE)\s+(\d+):\s*$/) {
                $objtype = '';
                $hoyde = '';
                $arealtype = '';
                $artype = '';
                $objnr = $2;
            }
            elsif (/\.\.OBJTYPE (.+)\s*$/) {
                $objtype = $1;
                $objtype =~ tr//aoaAOA/;
            }
            elsif (/\.\.TEGNSETT\s+UTF-8/) {
                $UTF8_TO_ANSI = 1;
                printf(STDERR "Input file is using UTF-8!\n");
            }		
            elsif (/\.\.\.KOORDSYS\s+(\d+)/) {
                $SOURCEUTM = sprintf("%dN", $1+10);
                printf(STDERR "Source UTM zone = $SOURCEUTM\n");
				if ($TARGETUTM) {
					if ($TARGETUTM eq $SOURCEUTM) {
						printf(STDERR "Target UTM = Source UTM so no conversion!\n");
						#$TARGETUTM = '';
						#$SOURCEUTM = '';
					}
					else {
						printf("Converting from %s to %s\n", $SOURCEUTM, $TARGETUTM);
					}
				}
            }		
            elsif (/\.\.ARTYPE (.+)\s*$/) {
                $artype = $1;
                if (defined($ArType{$artype}) && $objtype eq 'ArealressursFlate') {
                    $objtype = 'ARF_'.$ArType{$artype};
                }
            }
            elsif (/\.\.HYDE (.+)\s*$/) {
                $hoyde = $1;
            }
            elsif (/\.\.\.ENHET (.+)\s*$/) {
                $ENHET = $1;
                printf(STDERR "ENHET = $1\n");
            }
            elsif (/\.\.\.ORIGO-N\s+(\d+)\s+(\d+)\s*$/) {
                @ORIGO_NE = ($1, $2);
                printf(STDERR "Origo-N = $1,$2\n");
            }
            elsif (/\.DATAFANGSTDATO\s+(\d\d\d\d)(\d\d)(\d\d)/) {
                ($yy, $mm, $dd) = ($1, $2, $3);
                $datafangstdato = sprintf("%4d-%02d-%02d", $yy, $mm, $dd);
            }
            elsif (/\.VERIFISERINGSDATO\s+(\d\d\d\d)(\d\d)(\d\d)/) {
                ($yy, $mm, $dd) = ($1, $2, $3);
                $verifiseringsdato = sprintf("%4d-%02d-%02d", $yy, $mm, $dd);
            }
            elsif (/\.KOPIDATO\s+(\d\d\d\d)(\d\d)(\d\d)/) {
                ($yy, $mm, $dd) = ($1, $2, $3);
                $kopidato = sprintf("%4d-%02d-%02d", $yy, $mm, $dd);
            }
            elsif (/\.ARAVGRTYPE (.+)\s*$/) {
                $arealtype = $1;
            }
            elsif (/^\.FLATE\s+(\d+)\s*:/) {
                $flatenr = $1;
                #$nr = $2;
            }
            elsif (/^\.\.REF\s+:(.*)$/) {
                $ref = $1;
                while (1) {
                    $_ = <$sf>;
                    chomp;
                    last if (/^\./);
                    $ref .= $_;
                }
                $ref =~ s/\s+//g;
            }
            elsif (/^\.+(\w+)/) {
                $typ = $1;
                #$nr = $2;
            }
            undef $pe;
        }
        elsif (/^(\d+)\s+(\d+)/) {
            my ($e, $n, $h) = (conv($2,$1),$hoyde);
            if (/^(\d+)\s+(\d+)\s+(\d+)/) {
                ($e, $n, $h) = conv($2,$1,$3);
            }
			if ($TARGETUTM ne $SOURCEUTM) {
				die("Target UTM = $TARGETUTM, Source UTM is missing!") unless ($SOURCEUTM);
				my ($lat, $lon) = utm_to_latlon('wgs-84',$SOURCEUTM ,$e,$n);
				my $zone;
				($zone, $e, $n) = latlon_to_utm_force_zone('wgs-84', $TARGETUTM, $lat, $lon);
				$e = sprintf("%1.2f", $e); $n = sprintf("%1.2f", $n); 
			}
            #printf(STDERR "$1,$2 -> $n,$e\n");
            $h = '' unless defined($h);
            my ($x, $y) = ($e, $n);
            $xmin = $x if ($x < $xmin);
            $xmax = $x if ($x > $xmax);
            $ymin = $y if ($y < $ymin);
            $ymax = $y if ($y > $ymax);

            if (0 && defined($pe) && $pe) {
                my $de = $e-$pe;
                my $dn = $n-$pn;
                my $len2 = ($de*$de+$dn*$dn);
                if ($len2 >= 5000*5000) { # > 2000m!
                    printf("Ekstra flush at line $line ($objtype delta = %1.1f\n",sqrt($len2));
                    flushprev;
                }
            }
            $pe = $e; $pn = $n;
            push(@coords, "$e,$n,$h");
        }
    }
    flushprev;

    printf(DXF "0\nSECTION\n2\nHEADER\n9\n\$EXTMIN\n10\n%1.2f\n20\n%1.2f\n9\n\$EXTMAX\n10\n%1.2f\n20\n%1.2f\n0\nENDSEC\n",
            $xmin, $ymin, $xmax, $ymax);
    printf(DXF "0\nSECTION\n2\nENTITIES\n");

    my @banks = ();
    my @ebanks = ();

    my %flater = ();
    foreach (sort keys %objects) {
        my $obj = $_;
        my $nr = 0;
        printf(STDERR "$CID%d %s\n", scalar(@{$objects{$obj}}), $obj);
		next if (defined($ignore_sosi{$obj}));

        foreach (@{$objects{$obj}}) {
            $nr++;
            my @coords = split(/;/);
            
            if ($obj eq 'flate') {
                my $o = "flate_" . shift @coords;
				next if (defined($ignore_sosi{$o}));
				$seen_crt{$o}++;
                $objects{$o} = 1 unless defined($objects{$o}); # Force CRT entry!
                unless (defined($flater{$o})) {
                    printf(STDERR "Ny flate type: %s\n", $o);
                }
                $flater{$o}++;
                $ref = shift @coords;
                my ($x, $y, $h) = split(/,/, $coords[0]);
                my @holes = ();
                @coords = ();
                while ($ref =~ /\(([^\)]*)\)/) {
                    push(@holes, $1);
                    $ref = $`.$';
                }
                @coords = ref2coord($ref);
                $h = '' unless defined($h);
                if ($h =~ /\d/) {
                    $h = "38\n$h\n";
                }
				if (scalar(@coords) < 3) {
					printf(STDERR "$CID Flate med for f (%d) punkter!\n", scalar(@coords));
					next;
				}
                if ($coords[0] ne $coords[-1]) {
					#printf(STDERR "%s <=> %s\n", $coords[0], $coords[-1]);
					my $dist = dist2d($coords[0], $coords[-1]);
					if ($dist > 50) {
						printf(STDERR "$CID Flate %s har et hull av lengde %f, hopper over!\n", $o, $dist);
						next;
					}
                    printf(STDERR "$CID Flate %s som ikke er lukket (dist = %f), fixing it...\n", $o, $dist) if ($dist > 0.0);
					push(@coords, $coords[0]) if ($dist > 0.0);
                }
                
                #push(@coords, $coords[0]); # Should not be needed!
				
				my $inside = 0;
				# Include 92 / 2 parameter from Sverre Foyen:
                my $dxfs = sprintf("  0\nHATCH\n 5\n27\n100\nAcDbEntity\n  8\n%s\n%s100\nAcDbEntity\n  2\nSOLID\n 70\n1\n 91\n2\n 92\n2\n 73\n1\n 93\n%d\n",
                    $o, $h, scalar(@coords));
                foreach (@coords) {
                    ($x, $y, $h) = split(/,/);
					if (!$inside) {
						$inside = ($x >= $XMIN && $x < $XMAX && $y >= $YMIN && $y < $YMAX);
					}
                    $dxfs .= sprintf("10\n%1.2f\n20\n%1.2f\n", $x, $y);
                }
                if ($inside && scalar(@holes)) {
                    foreach (@holes) {
                        my @points = ref2coord($_);
                        $dxfs .= sprintf(" 73\n1\n 93\n%d\n", scalar(@points));
                        foreach (@points) {
                            ($x, $y, $h) = split(/,/);
                            $dxfs .= sprintf("10\n%1.2f\n20\n%1.2f\n", $x, $y);
                        }
                    }
                }
				printf( DXF "%s", $dxfs) if ($inside);
				#printf( STDERR "s") unless ($inside);
                next;
            }
            
            # Object must be point or linear
            #my $ymd = shift @coords;
			# Hack to filter out duplicate coords for Oslo 'Mast' objects:
			if ((scalar(@coords) == 2) && ($coords[0] eq $coords[1])) {
				shift @coords;
			}
            if (scalar(@coords) > 1) {
                #if ($obj ne 'Innsjo' && $obj =~ /Innsjo/) {	# Innsj/vann kanter!
                    #push(@banks, $_);
                #}
                #if ($obj =~ /Elv/ && int($crt{$obj})) {	# Elvekanter
                #    push(@ebanks, $_);
                #}
                #if ($obj eq 'KantUtsnitt') {
                    #push(@banks, $_);
                    #push(@ebanks, $_);
                #}
                my ($x, $y, $h) = split(/,/, $coords[0]);
                
                if ($obj eq 'Hoydekurve' || $obj eq 'Tellekurve' || $obj eq 'Hjelpekurve') {
                    my $h1 = sprintf("%1.0f", $h);
					$obj = 'Hjelpekurve';
                    if (( $h1 % $EKVIDISTANSE) == 0) {
                        $obj = 'Hoydekurve';
                        $obj = 'Tellekurve' if (( $h1 % $TELLEKURVEEKVIDISTANSE) == 0);
                    }
                    else {
                        
                    }
                }
                $h = '' unless defined($h);
                if ($h =~ /\d/) {
                    $h = "38\n$h\n";
                }
				my $inside = 0;
                my $dxfs = sprintf("0\nLWPOLYLINE\n8\n%s\n%s90\n%d\n", $obj, $h, scalar(@coords));
                foreach (@coords) {
                    ($x, $y, $h) = split(/,/);
					if (!$inside) {
						$inside = ($x >= $XMIN && $x < $XMAX && $y >= $YMIN && $y < $YMAX);
					}
                    $dxfs .= sprintf("10\n%1.2f\n20\n%1.2f\n", $x, $y);
                }
                printf( DXF "%s0\nSEQEND\n", $dxfs) if ($inside);
				#printf( STDERR "S") unless ($inside);
            }
            elsif (scalar(@coords)) {
                my ($x, $y, $h) = split(/,/, $coords[0]);
				my $inside = ($x >= $XMIN && $x < $XMAX && $y >= $YMIN && $y < $YMAX);
                $h = '' unless defined($h);
                if ($h =~ /\d/) {
                    $h = "30\n$h\n";
                }
                printf( DXF "0\nPOINT\n8\n%s\n10\n%1.2f\n20\n%1.2f\n%s0\nSEQEND\n", $obj, $x, $y, $h) if ($inside);
				#printf( STDERR "Skipping point\n") unless ($inside);
            }
        }
    }

    printf( DXF "0\nENDSEC\n0\nEOF\n");
    close(DXF);

	my $crt = substr($dxf,0,length($dxf)-3).'crt';
	if ($CRT) {
		open( CRT, ">", $crt) || die("Could not write to $crt\n");
		$seen_crt{'Tellekurve'} = $seen_crt{'Hoydekurve'} = $seen_crt{'Hjelpekurve'} = 1;

		foreach (sort keys %seen_crt) {
			if (defined($crt{$_})) {
				printf( CRT "%s %s\n", $crt{$_}, $_);
			}
			else {
				printf(STDERR "'%s' => 0,\n", $_);
			}
		}
		foreach (sort keys %crt) {
			if (!defined($seen_crt{$_})) {
				printf( CRT "%s %s\n", $crt{$_}, $_);
			}
		}
		close(CRT);
	}

    printf(STDERR "$CID $dxf %screated!\n", $CRT ? "and $crt " : '');
}

sub ref2coord
{
	my ($ref) = @_;
	return () unless defined($ref);
	#$ref = s/\s//g;
	my @coords = ();
	foreach (split(/:/, $ref)) {
		next unless (/\d/);
		my $rev = 0;
		if ($_ < 0) {
			$rev = 1;
			$_ = -$_;
		}
		unless (defined($objlist{$_})) {
			printf(STDERR "Current ref = '%s'\n", $_);
		}
		else {
			my @c = @{$objlist{$_}};
			if ($rev) {
				@c = reverse @c;
			}
			if (scalar(@coords) && $coords[-1] eq $c[0]) {
				shift @c;  # Drop first (duplicate!) point
			}
			push(@coords, @c);
		}
	}
	return @coords;
}

sub square_dist2d
{
	my ($a, $b) = @_;
	my ($x0, $y0, $h0) = split(/,/, $a);
	my ($x1, $y1, $h1) = split(/,/, $b);
	my ($dx, $dy) = ($x1-$x0, $y1-$y0);
	return ($dx*$dx + $dy*$dy);
}

sub dist2d
{
	my ($a, $b) = @_;
	my ($x0, $y0, $h0) = split(/,/, $a);
	my ($x1, $y1, $h1) = split(/,/, $b);
	my ($dx, $dy) = ($x1-$x0, $y1-$y0);
	return sqrt($dx*$dx + $dy*$dy);
}

sub myglob
{
	my @match = ();
	foreach (@_) {
		my $dir = '.';
		if (/^(.+)[\\\/]([^\\\/]+)$/) { # Starts with a directory!
			$dir = $1;
			$_ = $2;
		}
		s/\./\t/g;		# Save away any explicit '.' symbols
		s/\*/\.*/g;		# Convert * to .*
		s/\?/\./g;		# Convert ? to .
		s/\t/\\\./g;	# Restore explicit '.' as escaped '\.'
		
		my $pattern = $_;
		
		opendir(D, $dir);
		my @files = readdir(D);
		closedir(D);
		
		$dir = ($dir eq '.') ? '' : $dir .'/';
		
		foreach (@files) {
			next unless (-f $dir.$_);
			
			next unless (/^$pattern$/);
			push(@match, $dir . $_);
		}
	}
	return @match;
}



my $tail = q(
sub key
{
	my ($p) = @_;
	my ($x, $y, $h) = split(/,/, $p);
	$p = sprintf("%1.0f,%1.0f", $x*$DIST_SCALE, $y*$DIST_SCALE);
	return $p;
}

sub akeys
{
	my ($p) = @_;
	my ($xs, $ys, $h) = split(/,/, $p);
	
	# Locate matches within the nearest 1/$DIST_SCALE meters:
	($xs, $ys) = (int($xs*$DIST_SCALE), int($ys*$DIST_SCALE));

	my @xs = ($xs, $xs+1);
	my @ys = ($ys, $ys+1);

	my @p = ();
	foreach (@ys) {
		my $y = $_;
		foreach (@xs) {
			my $x = $_;
			$p = sprintf("%1.0f,%1.0f", $x, $y);
			push(@p, $p);
		}
	}
	return @p;
}

sub is_loop
{
	my @points = @_;
	my $key = key($points[0]);
	foreach (akeys($points[-1])) {
		if ($_ eq $key) {
			return 1;
		}
	}
	return 0;
}
	
sub merge_lines
{
	my @contours = @_;
	
	printf(STDERR "\n%d lines before merging\n", scalar(@contours));
	my %heads = ();
	my %tails = ();
	my @loops = ();
	
	for (my $c = 0; $c < scalar(@contours); $c++) {
		my @points = split(/;/, $contours[$c]);
		# Is this a loop? If so, skip!
		if (is_loop(@points)) {
			push(@loops, $contours[$c]);
			undef $contours[$c];
			next;
		}
		
		my ($head, $tail, $found);
		
		do {
			$head = key($points[0]);
			$tail = key($points[-1]);
			$found = 0;
			
			# Do we need to reverse this segment?
			if (defined($heads{$head}) || defined($tails{$tail})) {
				@points = reverse @points;
				my $t = $tail;
				$tail = $head;
				$head = $t;
				# Will match in one of the two next tests!
			}
			if (defined($heads{$tail})) {
				# This line is in front of an existing one:
				my $h = $heads{$tail};
				my @p = split(/;/, $contours[$h]);
				# Use dist instead! die("No match!\n") unless ($tail eq key($p[0]));
				
				# Remove all traces of previous segment:
				undef $contours[$h];
				foreach (akeys($p[0])) {
					$heads{$_} = undef;
				}
				foreach (akeys($p[-1])) {
					$tails{$_} = undef;
				}
				
				shift @p; # Get rid of the extra/identical boundary point!
				push(@points, @p);
				#$tail = key($points[-1]);
				
				$found++;
			}
			
			if (defined($tails{$head})) {
				# This line follows an existing one:
				my $h = $tails{$head};
				my @p = split(/;/, $contours[$h]);
				# dist instead! die("No match!\n") unless ($head eq key($p[-1]));
				
				undef $contours[$h];
				foreach (akeys($p[0])) {
					$heads{$_} = undef;
				}
				foreach (akeys($p[-1])) {
					$tails{$_} = undef;
				}
				
				shift @points;
				@points = (@p, @points);
				#$head = $points[0];
				
				$found++;
			} 
			#printf(STDERR "%d", $found) if ($found);
			#printf(STDERR "%s\r", ('-','\\','|','/')[$c & 3]) if ($tid);
		} while ($found);
		
		my $cont = join(";", @points);
		if (is_loop(@points)) {
			push(@loops, $cont);
			undef $contours[$c];
			next;
		}

		# Push this segment onto the head and tail lists:
		$contours[$c] = $cont;
		foreach (akeys($points[0])) {
			$heads{$_} = $c;
		}
		foreach (akeys($points[-1])) {
			$tails{$_} = $c;
		}
	}
	printf(STDERR "%d loops after merging\n", scalar(@loops));
	my @c = ();
	foreach (@contours) {
		if (defined($_)) {
			push(@c, $_);
		}
	}
	push(@extra, @c);
	return (@loops); # Return only the loops!
}

sub inside
{
	my ($x0,$y0,$x1, $y1, $bank, $candidate) = @_;
	my @banks = split(/;/, $bank);
	my @candidate = split(/;/, $candidate);
	
	return 0 if (scalar(@banks) < 3 || scalar(@candidate) < 3);
	
	my ($sx, $sy) = split(/,/, $candidate[0]);
	return 0 if ($sx < $x0 || $sx >= $x1 || $sy < $y0 || $sy >= $y1);
	
	my ($x, $y);
	my $crossings = 0;
	foreach (@banks) {
		my ($px, $py) = ($x, $y);
		($x, $y) = split(/,/);
		next unless (defined($py));
		
		# Must cross the Y coordinate!
		next if ($y >= $sy && $py >= $sy);
		next if ($y < $sy && $py < $sy);
		
		# X coord must be <=
		next if ($px > $sx && $x > $sx);
		
		# Possible crossing to the left!
		my $ix = $px + ($x-$px) * ($sy-$py) / ($y - $py);
		if ($ix <= $sx) {
			$crossings++;
		}
	}
	return $crossings & 1;
}
);